#define STRICT
#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* USAGE: huffmake < huffcode > huffcode.out */

FILE* in;
FILE* out;

char line[1024];

typedef struct _NODE {
	short next[2];
} NODE;
NODE nodes[8192];
int node_cnt;

typedef struct _HUFF {
	int i_table;
	int linbits;
} HUFF;
HUFF huff[32];

void readline();
void run();

int
main(int argc, char* argv[])
{
	in = stdin;
	if(argc >= 1 + 1) {
		in = fopen(argv[1], "r");
		if(!in) abort();
	}

	out = stdout;
	if(argc >= 1 + 2) {
		out = fopen(argv[2], "w");
		if(!out) abort();
	}

	run();

	if(in != stdin) fclose(in);
	if(out != stdout) fclose(out);

	fprintf(stderr, "OK\n");
	Sleep(500);
	return 0;
}

void
readline()
{
	int i;

	do {
		if(!fgets(line, sizeof line, in)) abort();
	} while(line[0] == '#' || line[0] == '\n');

	i = strlen(line) - 1;
	if(line[i] != '\n') abort();
	line[i] = '\0';
}

void
run()
{
	int retval;
	int i_table, ii_table, cx, cy, linbits;
	int i_x, i_y, x, y, hlen;
	int i_node, i_hlen, bit;
	char hcod[100];

	for(i_table = 0; i_table < 32 ; i_table++) {
		readline();

		retval = sscanf(line, ".table%d%d%d%d",
			&ii_table, &cx, &cy, &linbits);
		if(retval != 4) abort();
		if(ii_table != i_table) abort();

		if(cx == 0 || cy == 0) continue;

		readline();
		retval = sscanf(line, ".reference%d",
			&ii_table);
		if(retval == 1) {
			huff[i_table].i_table = ii_table;
			huff[i_table].linbits = linbits;
			continue;
		}

		huff[i_table].i_table = i_table;
		huff[i_table].linbits = linbits;

		memset(nodes, 0, sizeof nodes);
		node_cnt = 1;
		for(i_x = 0; i_x < cx; i_x++) {
			for(i_y = 0; i_y < cy; i_y++) {
				if(i_x || i_y) readline();
				retval = sscanf(line, "%d%d%d%s", &x, &y, &hlen, hcod);
				if(retval != 4) abort();
				if(x != i_x) abort();
				if(y != i_y) abort();
				if(hlen != (int)strlen(hcod)) abort();

				i_node = 0;
				for(i_hlen = 0; i_hlen < hlen - 1; i_hlen++) {
					bit = hcod[i_hlen] - '0';
					if(!(bit == 0 || bit == 1)) abort();
					if(!nodes[i_node].next[bit]) {
						nodes[i_node].next[bit] = node_cnt++;
					}
					i_node = nodes[i_node].next[bit];
				}
				bit = hcod[i_hlen] - '0';
				if(!(bit == 0 || bit == 1)) abort();
				nodes[i_node].next[bit] = -(x << 4 | y);
			}
		}

		fprintf(out, "static const short huffman_table%d[][2] = {\n", i_table);
		for(i_node = 0; i_node < node_cnt; i_node++) {
			fprintf(out, "\t/*%3d*/\t{%4d,%4d},\n", i_node, nodes[i_node].next[0], nodes[i_node].next[1]);
		}
		fprintf(out, "};\n");
	}

	fprintf(out, "/* typedef struct _HUFFMAN_TABLE {\n");
	fprintf(out, " * 	const short (*table)[2];\n");
	fprintf(out, " * 	short linbits;\n");
	fprintf(out, " * } HUFFMAN_TABLE;\n");
	fprintf(out, " * extern const HUFFMAN_TABLE huffman_table[32];\n");
	fprintf(out, " */\n");
	fprintf(out, "const HUFFMAN_TABLE huffman_table[32] = {\n");
	for(i_table = 0; i_table < 32; i_table++) {
		if(!huff[i_table].i_table) {
			fprintf(out, "\t/*%2d*/\t{NULL,0},\n", i_table);
		} else {
			fprintf(out, "\t/*%2d*/\t{huffman_table%d%s,%2d},\n", i_table, huff[i_table].i_table, huff[i_table].i_table < 10 ? " " : "", huff[i_table].linbits);
		}
	}
	fprintf(out, "};\n");
}
